//  IoT電子工作キット sLab-Remo2
//  6-4.カメラ付スマートリモコン【屋外対応】

// ライブラリ、Configファイルを読み込み
#include <WiFiClientSecure.h>
#include <EEPROM.h>
#include <SPIFFS.h>
#include <ESPAsyncWebServer.h>
#include <PubSubClient.h>
#include <ArduinoJson.h>      // Versionは「5.13.5」
#include <Wire.h>
#include "WebSocketsServer.h"
#include "OV7670.h"
#include "config.h"

// 利用ライブラリの設定を変更（PubSubClientライブラリは128バイトなのを拡張）
#define MQTT_MAX_PACKET_SIZE 1024

// サーバ、クライアントなど各種設定
WiFiClientSecure wifiClient;                            // MQTT用
PubSubClient mqttClient(host, 8883, wifiClient);        // MQTT用
OV7670 cam;                                             // Camera用
AsyncWebServer webServer(webport);                      // Webサーバ
WebSocketsServer webSocket = WebSocketsServer(wsport);  // WebSocketサーバ

// 起動時、最初に実行される処理
void setup(void) {
  // Serial設定
  Serial.begin(115200);
  Serial.println("sLab-Remo2 Sart");
  // SPIFFS開始
  SPIFFS.begin();
  // EEPROM開始（サイズ指定：ボタン名10個分[65*10],Login情報[1001-1128]）
  EEPROM.begin(1500);
  // Wire開始、端子設定
  Wire.begin();
  Wire.setClock(400000);
  pinMode ( LED_PIN,  OUTPUT );
  pinMode ( IR_R_PIN, INPUT );
  pinMode ( IR_S_PIN, OUTPUT );
  // Wi-Fi接続処理
  setup_wifi();

  // カメラ初期化処理
  Serial.println( "cam init" );  
  esp_err_t err = cam.init(&cam_conf, CAM_RES, RGB565);
  if(err != ESP_OK){
    Serial.print( "cam.init ERR:" );
    Serial.println(err);
    // 失敗時は15秒後に再起動しRetry
    delay(15000);
    ESP.restart();
  }
  cam.vflip( false );

  // WebServer設定処理
  setup_webserver();
  // WebSocket起動
  webSocket.begin();
  webSocket.onEvent(onWebSocketEvent);
  Serial.println("##### setup completed #####");
}

// 無線Wi-Fi設定
void setup_wifi() {
  // 無線Wi-Fi情報設定
  Serial.println ( "Wi-Fi SetUp" );
  WiFi.config( ip, gateway, subnet, DNS );  // DNSがないとMQTT接続不可
  WiFi.begin ( ssid, password );
  // Wi-Fi接続処理（接続するまで無限ループ）
  while ( WiFi.status() != WL_CONNECTED ) {
    // LEDを１秒毎に点滅する
    ledFlag = !ledFlag;
    digitalWrite(LED_PIN, ledFlag);
    // １秒間Wait
    delay ( 1000 );
    Serial.print ( "." );
  }
  // Wi-Fi接続できたのでシリアルモニターにIPアドレス表示
  Serial.print ( "Wi-Fi Connected! IP address: " );
  Serial.println ( WiFi.localIP() );
  // LED点灯（Wi-Fi接続状態）
  digitalWrite ( LED_PIN, true );
  // ------- MQTT_CONNECTがONの場合のみ クライアントにCA証明書を設定-------
  if ( MQTT_CONNECT ) {
    wifiClient.setCACert(beebottle_ca_cert);
  }
  // NTPサーバ設定
  configTime(9 * 3600L, 0, "ntp.nict.jp", "time.google.com", "ntp.jst.mfeed.ad.jp");
  // DDNS登録
  if (DDNS_CONNECT) {
    ddnsUp();
  }
}

// setup完了後、電源OFFまで繰り返し処理
void loop(void){
  // ------- 現在時刻の取得 -------
  nowtime = millis();
  // ------- カメラ映像処理 -------
  webSocket.loop();
  // WebSocket接続判定
  if (wsFlag) {
    logintime = nowtime;
    // カメラデータ読み込み処理
    cam.getLines( 1 , wsCamData , CAM_HEIGHT);
    // 映像データ送信処理
    webSocket.sendBIN(clientNum, wsCamData, dataSize);
    //Serial.println("ws.send");
  }
  // ------- MQTT_CONNECTがONの場合のみ -------
  if ( MQTT_CONNECT ) {
    // MQTT状態を確認し接続状態でなかったらMQTT接続処理
    if ( !mqttClient.connected() ) {
      reconnect();
    }
    // MQTTクライアント処理
    mqttClient.loop();
  }
  // ------- DDNS_CONNECTがONの場合のみ -------
  if (DDNS_CONNECT) {
    // DDNS登録用(10分周期で登録)
    if ( (int)(nowtime / 600000) != dCheck ) {
      dCheck = (int)(nowtime / 600000);
      ddnsUp();
    }
  }
  // ------- 自動Logout処理　-------
  if ( loginFlag && ( (nowtime - logintime) > (logTime * 1000) ) ) {
    loginFlag = false;
    loginIp = "";
    Serial.println("Auto Logout:" + String(nowtime) + ":" + String(logintime) );
    Serial.println(logintime);
  }
}

// MQTT接続処理
void reconnect() {
  // MQTT接続状態までループ処理
  while (!mqttClient.connected()) {
    Serial.println("Attempting MQTT connection...");
    // MQTT設定情報定義
    String username = "token:";
    username += channelToken;
    // MQTT接続処理
    mqttClient.connect(clientID, username.c_str(), NULL);
    delay(2000);
  }
  Serial.println("MQTT connected");
  // MQTTメッセージ受信時の処理設定
  mqttClient.setCallback(callback);
  // MQTTメッセージを受信するTOPICの設定
  mqttClient.subscribe(topic);
}

// MQTTメッセージ受信時の処理
void callback(char* topic, byte* payload, unsigned int length) {
  // MQTT受信メッセージを変数に保存
  char recvData[MQTT_MAX_PACKET_SIZE];
  snprintf(recvData, sizeof(recvData), "%s", payload);
  // MQTT受信メッセージをシリアルモニタ表示
  Serial.print("Message arrived [");
  Serial.print(topic);
  Serial.println("] ");
  Serial.println(recvData);
  // 受信データのJSON形式を解析し変数へ保存
  StaticJsonBuffer<MQTT_MAX_PACKET_SIZE> jsonBuffer;
  JsonObject& jsonBuf = jsonBuffer.parseObject(recvData);
  // JSON形式解析が成功しなかったらエラー表示し終了
  if (!jsonBuf.success()) {
    Serial.println("parseObject() failed");
    return;
  }
  // 受信データ（data）を取得保存
  const char* parsedPayload = jsonBuf["data"];
  // 受信データ（data）が存在するか判定
  if (parsedPayload != NULL) {
    Serial.print("payload: ");
    Serial.println(parsedPayload);
    // 受信データ（data）に送信リモコン番号
    if (contRemocon( parsedPayload )) {
      Serial.println("MQTT send OK");
    } else {
      Serial.println("MQTT send NG");
    }
  }
}
